@@ -49,7 +49,7 @@ static int win_fd;
 static int pty_fd = -1;
 static int log_fd = -1;
 #if CYGWIN_VERSION_API_MINOR >= 74
-static struct winsize prev_winsize = (struct winsize){0, 0, 0, 0};
+static struct winsize prev_winsize = {0, 0, 0, 0};
 #else
 static struct winsize prev_winsize;
 #endif
@@ -278,6 +278,22 @@ trim_environment(void)
   trimenv("BYOBU_");
 }
 
+static bool
+ispathprefix(string pref, string path)
+{
+  if (*pref == '/')
+    pref++;
+  if (*path == '/')
+    path++;
+  int len = strlen(pref);
+  if (0 == strncmp(pref, path, len)) {
+    path += len;
+    if (!*path || *path == '/')
+      return true;
+  }
+  return false;
+}
+
 void
 child_create(char *argv[], struct winsize *winp)
 {
@@ -294,6 +310,19 @@ child_create(char *argv[], struct winsize *winp)
   signal(SIGTERM, sigexit);
   signal(SIGQUIT, sigexit);
 
+  // support OSC 7 directory cloning if cloning WSL window while in rootfs
+  if (support_wsl && wslname) {
+    // this is done once in a new window, so don't care about memory leaks
+    char * rootfs = path_win_w_to_posix(wsl_basepath);
+    char * cwd = getcwd(malloc(MAX_PATH), MAX_PATH);
+    if (ispathprefix(rootfs, cwd)) {
+      char * dist = cs__wcstombs(wslname);
+      char * wslsubdir = cwd + strlen(rootfs);
+      char * wsldir = asform("//wsl$/%s%s", dist, wslsubdir);
+      chdir(wsldir);
+    }
+  }
+
   // Create the child process and pseudo terminal.
   pid = forkpty(&pty_fd, 0, 0, winp);
   if (pid < 0) {
@@ -648,7 +677,7 @@ child_kill(bool point_blank)
 bool
 child_is_alive(void)
 {
-    return pid;
+  return pid;
 }
 
 bool
