@@ -19,6 +19,46 @@
 #include <sys/cygwin.h>  // cygwin_internal
 #endif
 
+
+#ifdef check_cygdrive
+// adapt /cygdrive prefix to actual configured one (like / or anything else);
+// this experimental approach is not enabled as it is 
+// only used for WSL and dynamic adaptation is hardly useful
+
+static char * _cygdrive = 0;
+static wchar * _wcygdrive = 0;
+
+static char *
+cygdrive(void)
+{
+  if (!_cygdrive) {
+    char target [99];
+    int ret = readlink ("/proc/cygdrive", target, sizeof (target) - 1);
+    if (ret >= 0) {
+      target [ret] = '\0';
+      _cygdrive = strdup(target);
+    }
+    else
+#if defined(__MSYS__) || defined(__MINGW32)
+      _cygdrive = "/";
+#else
+      _cygdrive = "/cygdrive";
+#endif
+    _wcygdrive = cs__mbstowcs(target);
+  }
+  return _cygdrive;
+}
+
+static wchar *
+wcygdrive(void)
+{
+  (void)cygdrive();
+  return _wcygdrive;
+}
+
+#endif
+
+
 static DWORD WINAPI
 shell_exec_thread(void *data)
 {
@@ -57,7 +97,7 @@ shell_exec(wstring wpath)
   // this frees wpath via shell_exec_thread
 }
 
-#define dont_debug_wsl
+#define dont_debug_wslpath 3
 
 static struct {
   string p;
@@ -112,11 +152,102 @@ static int wsl_fstab_len = 0;
 static char *
 skip(char * s)
 {
-  while (isspace(*s))
+  while (iswspace(*s))
     s++;
   return s;
 }
 
+static int
+pathpref(char * x, char * y)
+{
+  char _x = *x;
+  char _y = *y;
+  if (!_x) {
+    if (!_y)
+      return 2;  // equal
+    else
+      return 1;  // prefix
+  }
+  if (_x == '\\')
+    _x = '/';
+  if (_y == '\\')
+    _y = '/';
+  if (_x != _y)
+    return 0;
+  return pathpref(x + 1, y + 1);
+}
+
+#if 0
+static char *
+strsubst(char * str, char c, char d)
+{
+  char * s = str;
+  while ((s = strchr(s, c)))
+    *s = d;
+  return str;
+}
+#endif
+
+#ifdef debug_wslpath
+static inline void
+show_fstab(void)
+{
+#if defined(debug_wslpath) && debug_wslpath > 2
+  for (int i = 0; i < wsl_fstab_len; i++)
+    printf("wslpath[%d] <%s> <%s>\n", i, wsl_fstab[i].dev_fs, wsl_fstab[i].mount_point);
+#endif
+}
+#endif
+
+static void
+windrives(void)
+{
+  HANDLE hnet;
+  if (WNetOpenEnum(RESOURCE_CONNECTED, RESOURCETYPE_DISK, 0, NULL, &hnet))
+    return;
+
+  DWORD bufsiz = 16384;
+  NETRESOURCEW * nr = malloc(bufsiz);
+  for (;;) {
+    //int res = WNetEnumResourceW(hnet, &(DWORD){1}, nr, &bufsiz);
+    DWORD nrn = 1;
+    int res = WNetEnumResourceW(hnet, &nrn, nr, &bufsiz);
+    if (res == NO_ERROR || res == ERROR_MORE_DATA) {
+#if defined(debug_wslpath) && debug_wslpath > 1
+      printf("net use <%ls> <%ls>\n", nr->lpLocalName, nr->lpRemoteName);
+#endif
+      if (!nr->lpLocalName)
+        continue;  // skip false matches on UNC path net mounts
+
+      // we accept some initial memory leak of configuration strings here
+      char * localdrive = cs__wcstoutf(nr->lpLocalName);
+      char * remoteshare = cs__wcstoutf(nr->lpRemoteName);
+      // adjust fstab with {"X:", "\\host\share"}
+      for (int i = 0; i < wsl_fstab_len; i++) {
+        // for {"X:", "\\host\share"},
+        // and all wsl_fstab entries 
+        // {"//host/share", "/mnt/..."} or
+        // {"\\host\share", "/mnt/..."},
+        // replace "//host/share" with "X:",
+        // also with subdir entries like
+        // {"//host/share/subdir", "/mnt/..."}
+        // replace "//host/share/subdir" with "X:\subdir"
+        switch (pathpref(remoteshare, wsl_fstab[i].dev_fs)) {
+          when 2: wsl_fstab[i].dev_fs = localdrive;
+          when 1: {  // prefix
+                  char * subdir = wsl_fstab[i].dev_fs + strlen(remoteshare);
+                  wsl_fstab[i].dev_fs = asform("%s%s", localdrive, subdir);
+          }
+        }
+      }
+    }
+    else
+      break;
+  }
+
+  WNetCloseEnum(hnet);
+}
+
 static bool
 wslmntmapped(void)
 {
@@ -163,17 +294,20 @@ wslmntmapped(void)
         char * p1 = skip(linebuf);
         if (*p1 != '#') {
           char * x = p1;
-          while (!isspace(*x))
+          while (!iswspace(*x))
             x++;
           *x++ = 0;
           char * p2 = skip(x);
           x = p2;
-          while (!isspace(*x))
+          while (!iswspace(*x))
             x++;
           *x = 0;
           if (x-- > p2 && *x == '/')
             *x = 0;
           if (*p1 && *p2) {
+#if defined(debug_wslpath)
+            printf("fstab <%s> <%s>\n", p1, p2);
+#endif
             wsl_fstab = renewn(wsl_fstab, wsl_fstab_len + 1);
             wsl_fstab[wsl_fstab_len].dev_fs = strdup(p1);
             wsl_fstab[wsl_fstab_len].mount_point = strdup(p2);
@@ -188,6 +322,9 @@ wslmntmapped(void)
     free(wsl$conf);
     free(rootfs);
     wslmnt_init = true;
+
+    // substitute fstab dev_fs entries with Windows notation
+    windrives();
   }
   return wsl_conf_mnt || wsl_fstab;
 }
@@ -201,14 +338,32 @@ unwslpath(wchar * wpath)
     if (wsl_conf_mnt && ispathprefix(wsl_conf_mnt, wslpath))
       res = asform("/cygdrive%s", &wslpath[strlen(wsl_conf_mnt)]);
     else
-      for (int i = 0; i < wsl_fstab_len; i++)
-        if (ispathprefix(wsl_fstab[i].mount_point, wslpath))
+      for (int i = 0; i < wsl_fstab_len; i++) {
+#if defined(debug_wslpath) && debug_wslpath > 2
+        printf("unwslpath <%s> <%s> (%s)\n", wslpath, wsl_fstab[i].mount_point, wsl_fstab[i].dev_fs);
+#endif
+        if (ispathprefix(wsl_fstab[i].mount_point, wslpath)) {
           if (wsl_fstab[i].dev_fs[1] == ':') {
-            res = asform("/cygdrive/%c%s", 
-                         tolower(wsl_fstab[i].dev_fs[0]), 
+            res = asform(
+#if defined(__MSYS__) || defined(__MINGW32)
+#else
+                         "/cygdrive"
+#endif
+                         "/%c%s%s", 
+                         tolower((int)wsl_fstab[i].dev_fs[0]), 
+                         &wsl_fstab[i].dev_fs[2], 
                          &wslpath[strlen(wsl_fstab[i].mount_point)]);
             break;
           }
+          else {  // handle UNC path mount
+            res = asform("%s%s", wsl_fstab[i].dev_fs, wslpath + strlen(wsl_fstab[i].mount_point));
+            for (char * r = res; *r; r++)
+              // cygwin path conversion fails on backslash UNC paths
+              if (*r == '\\')
+                *r = '/';
+          }
+        }
+      }
     free(wslpath);
   }
   return res;
@@ -218,19 +373,57 @@ static char *
 wslpath(char * path)
 {
   char * res = null;
-  if (ispathprefix("/cygdrive", path) && wslmntmapped()) {
-    path += 10;  // point to drive letter
-    if (wsl_conf_mnt)
+  if (wslmntmapped() &&
+#if defined(__MSYS__) || defined(__MINGW32)
+      path[2] == '/'
+#else
+      ispathprefix("/cygdrive", path)
+#endif
+     )
+  {  // check drive path
+#if defined(__MSYS__) || defined(__MINGW32)
+#else
+    path += 9;
+#endif
+    if (*path == '/')
+      path++;  // point to drive letter
+    if (wsl_conf_mnt) {  // [automount] configured
       res = asform("%s/%s", wsl_conf_mnt, path);
-    else if (path[1] == '/')
-      for (int i = 0; i < wsl_fstab_len; i++)
-        if (tolower(*path) == tolower(wsl_fstab[i].dev_fs[0])
+    }
+    else if (path[1] == '/') {
+      for (int i = 0; i < wsl_fstab_len; i++) {
+#if defined(debug_wslpath) && debug_wslpath > 2
+        printf("wslpath <%s> fstab <%s> <%s>\n", path, wsl_fstab[i].dev_fs, wsl_fstab[i].mount_point);
+#endif
+        if (tolower((int)*path) == tolower((int)wsl_fstab[i].dev_fs[0])
          && wsl_fstab[i].dev_fs[1] == ':'
-         && wsl_fstab[i].dev_fs[2] == 0) {
-            path ++;
-            res = asform("%s%s", wsl_fstab[i].mount_point, path);
+           )
+        {
+          char * dev = &wsl_fstab[i].dev_fs[2];
+          if (strlen(dev) == 1) {
+            //path <x/dir/file> fstab <X:\> </mnt/mountpoint>
+            res = asform("%s%s", wsl_fstab[i].mount_point, path + 1);
             break;
           }
+          else if (pathpref(dev, path + 1)) {
+            //path <x/dir/file> fstab <X:\subdir> </mnt/mountpoint>
+            res = asform("%s%s", wsl_fstab[i].mount_point, path + 1 + strlen(dev));
+            break;
+          }
+        }
+      }
+    }
+  }
+  else if (wslmntmapped()) {  // check UNC path
+    for (int i = 0; i < wsl_fstab_len; i++) {
+#if defined(debug_wslpath) && debug_wslpath > 2
+      printf("wslpath <%s> <%s> (%s)\n", path, wsl_fstab[i].mount_point, wsl_fstab[i].dev_fs);
+#endif
+      if (pathpref(wsl_fstab[i].dev_fs, path)) {
+        res = asform("%s%s", wsl_fstab[i].mount_point, path + strlen(wsl_fstab[i].dev_fs));
+        break;
+      }
+    }
   }
   // cygwin-internal paths not supported
   return res;
@@ -239,8 +432,8 @@ wslpath(char * path)
 wchar *
 dewsl(wchar * wpath)
 {
-#ifdef debug_wsl
-  printf("open <%ls>\n", wpath);
+#ifdef debug_wslpath
+  printf("dewsl <%ls>\n", wpath);
 #endif
   char * unwslp = unwslpath(wpath);
   if (unwslp) {
@@ -357,8 +550,8 @@ win_open(wstring wpath, bool adjust_dir)
       wpath = dewsl((wchar *)wpath);
     }
     wstring conv_wpath = child_conv_path(wpath, adjust_dir);
-#ifdef debug_wsl
-    printf("open <%ls> <%ls>\n", wpath, conv_wpath);
+#ifdef debug_wslpath
+    printf("win_open <%ls> <%ls>\n", wpath, conv_wpath);
 #endif
     delete(wpath);
     if (conv_wpath)
@@ -425,6 +618,7 @@ win_copy(const wchar *data, cattr *cattrs, int len)
 void
 win_copy_as(const wchar *data, cattr *cattrs, int len, char what)
 {
+  //printf("win_copy %d '%c'\n", len, what);
   HGLOBAL clipdata, clipdata2, clipdata3 = 0;
   int len2;
   void *lock, *lock2, *lock3;
@@ -441,10 +635,17 @@ win_copy_as(const wchar *data, cattr *cattrs, int len, char what)
       GlobalFree(clipdata2);
     return;
   }
-  if (!(lock = GlobalLock(clipdata)))
+  if (!(lock = GlobalLock(clipdata))) {
+    GlobalFree(clipdata);
+    GlobalFree(clipdata2);
     return;
-  if (!(lock2 = GlobalLock(clipdata2)))
+  }
+  if (!(lock2 = GlobalLock(clipdata2))) {
+    GlobalUnlock(clipdata);
+    GlobalFree(clipdata);
+    GlobalFree(clipdata2);
     return;
+  }
 
   memcpy(lock, data, len * sizeof(wchar));
   WideCharToMultiByte(CP_ACP, 0, data, len, lock2, len2, null, null);
@@ -472,19 +673,21 @@ win_copy_as(const wchar *data, cattr *cattrs, int len, char what)
       MultiByteToWideChar(CP_ACP, MB_ERR_INVALID_CHARS,
                           (char[]){i}, 1, unitab + i, 1);
 
-    char * rtffontname = newn(char, wcslen(cfg.font.name) * 9 + 1);
+    wstring cfgfont = *cfg.copy_as_rtf_font ? cfg.copy_as_rtf_font : cfg.font.name;
+    int cfgsize = cfg.copy_as_rtf_font_size ? cfg.copy_as_rtf_font_size : cfg.font.size;
+    char * rtffontname = newn(char, wcslen(cfgfont) * 9 + 1);
     char * rtffnpoi = rtffontname;
-    for (uint i = 0; i < wcslen(cfg.font.name); i++)
-      if (!(cfg.font.name[i] & 0xFF80) && !strchr("\\;{}", cfg.font.name[i]))
-        *rtffnpoi++ = cfg.font.name[i];
+    for (uint i = 0; i < wcslen(cfgfont); i++)
+      if (!(cfgfont[i] & 0xFF80) && !strchr("\\;{}", cfgfont[i]))
+        *rtffnpoi++ = cfgfont[i];
       else
-        rtffnpoi += sprintf(rtffnpoi, "\\u%d '", cfg.font.name[i]);
+        rtffnpoi += sprintf(rtffnpoi, "\\u%d '", cfgfont[i]);
     *rtffnpoi = '\0';
     rtfsize = 100 + strlen(rtffontname);
     rtf = newn(char, rtfsize);
     rtflen = sprintf(rtf,
-      "{\\rtf1\\ansi\\deff0{\\fonttbl{\\f0\\fmodern %s;}}\\f0\\fs%d",
-      rtffontname, cfg.font.size * 2);
+      "{\\rtf1\\ansi\\deff0{\\fonttbl{\\f0\\fmodern\\fprq1 %s;}}\\f0\\fs%d",
+      rtffontname, cfgsize * 2);
     free(rtffontname);
 
    /*
@@ -700,25 +903,30 @@ win_copy_as(const wchar *data, cattr *cattrs, int len, char what)
   GlobalUnlock(clipdata);
   GlobalUnlock(clipdata2);
 
+  //printf("OpenClipboard win_copy\n");
   if (OpenClipboard(wnd)) {
     clipboard_token = true;
     EmptyClipboard();
+
     // copy clipboard text formats
     SetClipboardData(CF_UNICODETEXT, clipdata);
     SetClipboardData(CF_TEXT, clipdata2);
+
     // copy clipboard RTF format
     if (clipdata3)
       SetClipboardData(RegisterClipboardFormat(CF_RTF), clipdata3);
     // determine HTML format level requested
     int level = 0;
     if (cfg.copy_as_html && !what)
-      level = cfg.copy_as_rtf ? 2 : 3;
+      //level = cfg.copy_as_rtf ? 2 : 3;
+      level = cfg.copy_as_html;
     else if (what == 'h')
       level = 1;
     else if (what == 'f')
       level = 2;
     else if (what == 'H')
       level = 3;
+
     // copy clipboard HTML format
     UINT CF_HTML = level ? RegisterClipboardFormatA("HTML Format") : 0;
     if (CF_HTML) {
@@ -748,10 +956,59 @@ win_copy_as(const wchar *data, cattr *cattrs, int len, char what)
         free(htmlcb);
         GlobalUnlock(clipdatahtml);
         SetClipboardData(CF_HTML, clipdatahtml);
+      }
+      else {
         GlobalFree(clipdatahtml);
       }
     }
 
+#if CYGWIN_VERSION_API_MINOR >= 74
+    // copy clipboard cygwin format, including timestamp
+    UINT CF_CYGCB = RegisterClipboardFormatW(W("CYGWIN_NATIVE_CLIPBOARD"));
+    if (CF_CYGCB &&
+# ifdef __CYGWIN32__
+        cygver_ge(3, 3)  // unified native clipboard/timestamp format
+# else
+        cygver_ge(1, 8)  // actually 1.7.13, according to putclip source
+# endif
+       )
+    {
+      int lenc = WideCharToMultiByte(cs_get_codepage(), 0, data, len, 0, 0, null, null);
+      HGLOBAL clipdatac = GlobalAlloc(GMEM_DDESHARE | GMEM_MOVEABLE, lenc + 24);
+      if (!clipdatac) {
+        goto nocygcb;
+      }
+      void * lockc = GlobalLock(clipdatac);
+      if (!lockc) {
+        GlobalFree(clipdatac);
+        goto nocygcb;
+      }
+
+      char * textbuf = lockc;
+      textbuf += 24;
+      int buflen = WideCharToMultiByte(cs_get_codepage(), 0, data, len, textbuf, lenc, null, null);
+      // strip terminating NUL (#1132)
+      if (!textbuf[buflen - 1])
+        buflen--;
+
+      struct {
+        u_int64_t cb_sec;
+        u_int64_t cb_nsec;
+        u_int64_t cb_size;
+      } * cygtime = lockc;
+      struct timespec timebuf;
+      clock_gettime (CLOCK_REALTIME, &timebuf);
+      cygtime->cb_nsec = timebuf.tv_nsec;
+      cygtime->cb_sec = timebuf.tv_sec;
+      cygtime->cb_size = buflen;
+
+      GlobalUnlock(clipdatac);
+      SetClipboardData(CF_CYGCB, clipdatac);
+
+      nocygcb:;
+    }
+#endif
+
     CloseClipboard();
   }
   else {
@@ -760,84 +1017,52 @@ win_copy_as(const wchar *data, cattr *cattrs, int len, char what)
   }
 }
 
-static char *
-matchconf(char * conf, char * item)
+static uint buf_len, buf_pos;
+static char * buf;
+
+static void buf_init()
 {
-  char * cmdp = conf;
-  char sepch = ';';
-  if ((uchar)*cmdp <= (uchar)' ')
-    sepch = *cmdp++;
-
-  char * paramp;
-  while ((paramp = strchr(cmdp, ':'))) {
-    *paramp = '\0';
-    paramp++;
-    char * sepp = strchr(paramp, sepch);
-    if (sepp)
-      *sepp = '\0';
-
-    if (!strcmp(cmdp, item))
-      return paramp;
-
-    if (sepp) {
-      cmdp = sepp + 1;
-      // check for multi-line separation
-      if (*cmdp == '\\' && cmdp[1] == '\n') {
-        cmdp += 2;
-        while (isspace(*cmdp))
-          cmdp++;
-      }
-    }
-    else
-      break;
-  }
-  return 0;
+  buf_len = 32;
+  buf_pos = 0;
+  buf = newn(char, buf_len);
 }
 
 static void
-paste_hdrop(HDROP drop)
+buf_add(char c)
 {
-  uint buf_len = 32, buf_pos = 0;
-  char *buf = newn(char, buf_len);
-  void buf_add(char c) {
-    if (buf_pos >= buf_len)
-      buf = renewn(buf, buf_len *= 2);
-    buf[buf_pos++] = c;
-  }
+  if (buf_pos >= buf_len)
+    buf = renewn(buf, buf_len *= 2);
+  buf[buf_pos++] = c;
+}
 
-#if CYGWIN_VERSION_API_MINOR >= 222
-  // Update Cygwin locale to terminal locale.
-  cygwin_internal(CW_INT_SETLOCALE);
-#endif
-  uint n = DragQueryFileW(drop, -1, 0, 0);
-  for (uint i = 0; i < n; i++) {
-
-    uint wfn_len = DragQueryFileW(drop, i, 0, 0);
-    wchar wfn[wfn_len + 1];
-    DragQueryFileW(drop, i, wfn, wfn_len + 1);
-#ifdef debug_dragndrop
-    printf("dropped file <%ls>\n", wfn);
-#endif
-    char *fn = path_win_w_to_posix(wfn);
+static void
+buf_path(wchar * wfn, bool convert, bool quote)
+{
+    bool posix_path = convert || support_wsl;
+    char * fn = posix_path
+              ? path_win_w_to_posix(wfn)
+              : cs__wcstoutf(wfn);
 
     bool has_tick = false, needs_quotes = false, needs_dollar = false;
     for (char *p = fn; *p && !needs_dollar; p++) {
       uchar c = *p;
       has_tick |= c == '\'';
-      needs_quotes |= isascii(c) && !isalnum(c) && !strchr("+,-./@_~'", c);
+      if (posix_path || !strchr("\\:", c))
+        needs_quotes |= isascii(c) && !isalnum(c) && !strchr("+,-./@_~'", c);
       needs_dollar = iscntrl(c) || (needs_quotes && has_tick);
     }
     needs_quotes |= needs_dollar;
+    needs_quotes &= quote;
 
     if (needs_dollar)
       buf_add('$');
     if (needs_quotes)
-      buf_add('\'');
+      buf_add(posix_path ? '\'' : '"');
     else if (*fn == '~')
       buf_add('\\');
     char *p = fn;
     if (support_wsl) {
-#ifdef debug_wsl
+#ifdef debug_wslpath
       printf("paste <%s>\n", p);
 #endif
       char * wslp = wslpath(p);
@@ -863,7 +1088,8 @@ paste_hdrop(HDROP drop)
         else if (strncmp(p, "/cygdrive/", 10) == 0) {
           // convert /cygdrive/X/path referring to mounted drive
           p += 5;
-          strncpy(p, "/mnt", 4);
+          //strncpy(p, "/mnt", 4);
+          memcpy(p, "/mnt", 4);
         }
       }
       else {
@@ -913,7 +1139,8 @@ paste_hdrop(HDROP drop)
           p = mp;
         else if (strncmp(p, "/cygdrive/", 10) == 0) {
           p += 5;
-          strncpy(p, "/mnt", 4);
+          //strncpy(p, "/mnt", 4);
+          memcpy(p, "/mnt", 4);
         }
       }
     }
@@ -932,11 +1159,34 @@ paste_hdrop(HDROP drop)
       }
     }
     if (needs_quotes)
-      buf_add('\'');
-    buf_add(' ');  // Filename separator
+      buf_add(posix_path ? '\'' : '"');
     free(fn);
+}
+
+static void
+paste_hdrop(HDROP drop)
+{
+#if CYGWIN_VERSION_API_MINOR >= 222
+  // Update Cygwin locale to terminal locale.
+  cygwin_internal(CW_INT_SETLOCALE);
+#endif
+  uint n = DragQueryFileW(drop, -1, 0, 0);
+
+  void bufpaths(bool convert, bool quote) {
+    buf_init();
+    for (uint i = 0; i < n; i++) {
+      uint wfn_len = DragQueryFileW(drop, i, 0, 0);
+      wchar wfn[wfn_len + 1];
+      DragQueryFileW(drop, i, wfn, wfn_len + 1);
+#ifdef debug_dragndrop
+      printf("dropped file <%ls>\n", wfn);
+#endif
+      if (i)
+        buf_add(' ');  // Filename separator
+      buf_path(wfn, convert, quote);
+    }
+    buf[buf_pos] = 0;
   }
-  buf_pos--;  // Drop trailing space
 
   if (!support_wsl && *cfg.drop_commands) {
     // try to determine foreground program
@@ -946,12 +1196,23 @@ paste_hdrop(HDROP drop)
       char * drops = cs__wcstombs(cfg.drop_commands);
       char * paste = matchconf(drops, fg_prog);
       if (paste) {
-        buf[buf_pos] = 0;
-        char * pastebuf = newn(char, strlen(paste) + strlen(buf) + 1);
-        sprintf(pastebuf, paste, buf);
-        child_send(pastebuf, strlen(pastebuf));
-        free(pastebuf);
-        free(drops);
+        char * format = strchr(paste, '%');
+        if (format && strchr("swSW", *(++format)) && !strchr(format, '%')) {
+          switch (*format) {
+            when 's': bufpaths(true, false);
+            when 'S': bufpaths(true, true);
+            when 'w': bufpaths(false, false);
+            when 'W': bufpaths(false, true);
+          }
+          *format = 's';
+          char * pastebuf = newn(char, strlen(paste) + strlen(buf) + 1);
+          sprintf(pastebuf, paste, buf);
+          child_send(pastebuf, strlen(pastebuf));
+          free(pastebuf);
+        }
+        else
+          child_send(paste, strlen(paste));
+        free(drops);  // also frees paste which points into drops
         free(fg_prog);
         free(buf);
         return;
@@ -961,6 +1222,24 @@ paste_hdrop(HDROP drop)
     }
   }
 
+  bufpaths(true, true);
+
+  if (term.bracketed_paste)
+    child_write("\e[200~", 6);
+  child_send(buf, buf_pos);
+  free(buf);
+  if (term.bracketed_paste)
+    child_write("\e[201~", 6);
+}
+
+static void
+paste_path(HANDLE data)
+{
+  wchar *s = GlobalLock(data);
+  buf_init();
+  buf_path(s, true, true);
+  GlobalUnlock(data);
+
   if (term.bracketed_paste)
     child_write("\e[200~", 6);
   child_send(buf, buf_pos);
@@ -989,9 +1268,10 @@ paste_text(HANDLE data)
   term_paste(s, l, (GetKeyState(VK_CONTROL) & 0x80) != 0);
 }
 
-void
-win_paste(void)
+static void
+do_win_paste(bool do_path)
 {
+  //printf("OpenClipboard win_paste\n");
   if (!OpenClipboard(null))
     return;
 
@@ -999,15 +1279,37 @@ win_paste(void)
     term.selected = false;
 
   HGLOBAL data;
-  if ((data = GetClipboardData(CF_HDROP)))
+  if ((data = GetClipboardData(CF_HDROP))) {
+    //printf("pasting CF_HDROP\n");
     paste_hdrop(data);
-  else if ((data = GetClipboardData(CF_UNICODETEXT)))
-    paste_unicode_text(data);
-  else if ((data = GetClipboardData(CF_TEXT)))
+  }
+  else if ((data = GetClipboardData(CF_UNICODETEXT))) {
+    //printf("pasting CF_UNICODETEXT\n");
+    if (do_path)
+      paste_path(data);
+    else
+      paste_unicode_text(data);
+  }
+  else if ((data = GetClipboardData(CF_TEXT))) {
+    //printf("pasting CF_TEXT\n");
     paste_text(data);
+  }
+
   CloseClipboard();
 }
 
+void
+win_paste(void)
+{
+  do_win_paste(false);
+}
+
+void
+win_paste_path(void)
+{
+  do_win_paste(true);
+}
+
 
 /*
  *  Drag-and-drop
